from .mailbox import MailBox
from .models import ServiceType, EmailMessage
from .errors import BetterImapException, IMAPLoginFailed, IMAPSearchTimeout
from .services import Service

__all__ = [
    "MailBox",
    "Service",
    "EmailMessage",
    "ServiceType",
    "BetterImapException",
    "IMAPLoginFailed",
    "IMAPSearchTimeout",
]



from .models import ServiceType
from enum import Enum


class Service:
    RAMBLER = ServiceType(
        name="Rambler",
        host="imap.rambler.ru",
        folders=["INBOX", "Spam"],
    )
    OUTLOOK = ServiceType(
        name="Outlook",
        host="outlook.office365.com",
        folders=["INBOX"],
    )
    ICLOUD = ServiceType(
        name="iCloud",
        host="imap.mail.me.com",
        folders=["INBOX"],
    )
    GMAIL = ServiceType(
        name="Gmail",
        host="imap.gmail.com",
        folders=["INBOX", "Spam"],
    )
    MAILRU = ServiceType(
        name="Mail.ru",
        host="imap.mail.ru",
        folders=["INBOX", "Spam"],
    )
    FIRSTMAIL = ServiceType(
        name="Firstmail",
        host="imap.firstmail.ltd",
        folders=["INBOX"],
    )


DOMAIN_TO_SERVICE = {
    "@rambler.ru": Service.RAMBLER,
    "@ro.ru": Service.RAMBLER,
    "@myrambler.ru": Service.RAMBLER,
    "@autorambler.ru": Service.RAMBLER,
    "@hotmail.com": Service.OUTLOOK,
    "@outlook.com": Service.OUTLOOK,
    "@icloud.com": Service.ICLOUD,
    "@gmail.com": Service.GMAIL,
    "@mail.ru": Service.MAILRU,
    "@inbox.ru": Service.MAILRU,
    "@bk.ru": Service.MAILRU,
    "@list.ru": Service.MAILRU,
    "@firstmail.ltd": Service.FIRSTMAIL,
    "@firstmail.ru": Service.FIRSTMAIL,
    "@nietamail.com": Service.FIRSTMAIL,
    "@menormail.com": Service.FIRSTMAIL,
    "@senoramail.com": Service.FIRSTMAIL,
    "@historiogramail.com": Service.FIRSTMAIL,
    "@ambismail.com": Service.FIRSTMAIL,
    "@andromomail.com": Service.FIRSTMAIL,
    "@superocomail.com": Service.FIRSTMAIL,
    "@velismail.com": Service.FIRSTMAIL,
    "@veridicalmail.com": Service.FIRSTMAIL,
    "@epidemiosmail.ru": Service.FIRSTMAIL,
    "@encepsmail.ru": Service.FIRSTMAIL,
    "@reevalmail.com": Service.FIRSTMAIL,
    "@decortiomail.ru": Service.FIRSTMAIL,
    "@decomposaomail.ru": Service.FIRSTMAIL,
    "@custoomail.ru": Service.FIRSTMAIL,
    "@curviomail.ru": Service.FIRSTMAIL,
    "@aberdoml.com": Service.FIRSTMAIL,
    "@adjudicaml.com": Service.FIRSTMAIL,
    "@admifml.com": Service.FIRSTMAIL,
    "@anathefml.com": Service.FIRSTMAIL,
    "@angulfml.com": Service.FIRSTMAIL,
    "@anticholineml.com": Service.FIRSTMAIL,
    "@apothegmaml.com": Service.FIRSTMAIL,
    "@archiepiscopaml.com": Service.FIRSTMAIL,
    "@autorotaml.com": Service.FIRSTMAIL,
    "@autotranfml.com": Service.FIRSTMAIL,
    "@bactfml.com": Service.FIRSTMAIL,
    "@barothermohygfml.com": Service.FIRSTMAIL,
    "@bimilleml.com": Service.FIRSTMAIL,
    "@biogeogfml.com": Service.FIRSTMAIL,
    "@biogeograpfml.com": Service.FIRSTMAIL,
    "@capabiml.com": Service.FIRSTMAIL,
    "@centralisaml.com": Service.FIRSTMAIL,
    "@cephafml.com": Service.FIRSTMAIL,
    "@churcfml.com": Service.FIRSTMAIL,
    "@commefml.com": Service.FIRSTMAIL,
    "@compromissoml.com": Service.FIRSTMAIL,
    "@consfml.com": Service.FIRSTMAIL,
    "@convolvufml.com": Service.FIRSTMAIL,
    "@corybantml.com": Service.FIRSTMAIL,
    "@counterrevfml.com": Service.FIRSTMAIL,
    "@deinstitutionalizaml.com": Service.FIRSTMAIL,
    "@depreciml.com": Service.FIRSTMAIL,
    "@deterioraml.com": Service.FIRSTMAIL,
    "@diacetonml.com": Service.FIRSTMAIL,
    "@diagonml.com": Service.FIRSTMAIL,
    "@diplofml.com": Service.FIRSTMAIL,
    "@discontinuml.com": Service.FIRSTMAIL,
    "@disoxygeml.com": Service.FIRSTMAIL,
    "@dissipatml.com": Service.FIRSTMAIL,
    "@documefml.com": Service.FIRSTMAIL,
    "@electrooptml.com": Service.FIRSTMAIL,
    "@epidemiogrml.com": Service.FIRSTMAIL,
    "@epithalaml.com": Service.FIRSTMAIL,
    "@euplexopml.com": Service.FIRSTMAIL,
    "@exentfml.com": Service.FIRSTMAIL,
    "@exheredaml.com": Service.FIRSTMAIL,
    "@expeditioml.com": Service.FIRSTMAIL,
    "@extefml.com": Service.FIRSTMAIL,
    "@factifml.com": Service.FIRSTMAIL,
    "@glacioloml.com": Service.FIRSTMAIL,
    "@hottentoml.com": Service.FIRSTMAIL,
    "@hydrofml.com": Service.FIRSTMAIL,
    "@hypodermicml.com": Service.FIRSTMAIL,
    "@ignoranml.com": Service.FIRSTMAIL,
    "@imagfml.com": Service.FIRSTMAIL,
    "@inanfml.com": Service.FIRSTMAIL,
    "@incunaml.com": Service.FIRSTMAIL,
    "@inexterfml.com": Service.FIRSTMAIL,
    "@inspiraml.com": Service.FIRSTMAIL,
    "@intercondyml.com": Service.FIRSTMAIL,
    "@interfml.com": Service.FIRSTMAIL,
    "@interpafml.com": Service.FIRSTMAIL,
    "@intersfml.com": Service.FIRSTMAIL,
    "@invariml.com": Service.FIRSTMAIL,
    "@irremfml.com": Service.FIRSTMAIL,
    "@kinesioml.com": Service.FIRSTMAIL,
    "@mastigophml.com": Service.FIRSTMAIL,
    "@menfml.com": Service.FIRSTMAIL,
    "@metathml.com": Service.FIRSTMAIL,
    "@methanomml.com": Service.FIRSTMAIL,
    "@mifml.com": Service.FIRSTMAIL,
    "@nectarivoml.com": Service.FIRSTMAIL,
    "@nonassessml.com": Service.FIRSTMAIL,
    "@noncontroverml.com": Service.FIRSTMAIL,
    "@nonefml.com": Service.FIRSTMAIL,
    "@nonrecogniml.com": Service.FIRSTMAIL,
    "@nonsubmfml.com": Service.FIRSTMAIL,
    "@occipitopariml.com": Service.FIRSTMAIL,
    "@opercfml.com": Service.FIRSTMAIL,
    "@opinionatiml.com": Service.FIRSTMAIL,
    "@oraculaml.com": Service.FIRSTMAIL,
    "@organophfml.com": Service.FIRSTMAIL,
    "@overdrafml.com": Service.FIRSTMAIL,
    "@overefml.com": Service.FIRSTMAIL,
    "@overstimulml.com": Service.FIRSTMAIL,
    "@palaeontofml.com": Service.FIRSTMAIL,
    "@paraprocml.com": Service.FIRSTMAIL,
    "@perinfml.com": Service.FIRSTMAIL,
    "@perissfml.com": Service.FIRSTMAIL,
    "@pharyngogloml.com": Service.FIRSTMAIL,
    "@physioloml.com": Service.FIRSTMAIL,
    "@politiciml.com": Service.FIRSTMAIL,
    "@polydactml.com": Service.FIRSTMAIL,
    "@proletarianizaml.com": Service.FIRSTMAIL,
    "@psychosexuaml.com": Service.FIRSTMAIL,
    "@psychosofml.com": Service.FIRSTMAIL,
    "@reprobatiml.com": Service.FIRSTMAIL,
    "@semidfml.com": Service.FIRSTMAIL,
    "@sitifml.com": Service.FIRSTMAIL,
    "@solecistml.com": Service.FIRSTMAIL,
    "@soliloquml.com": Service.FIRSTMAIL,
    "@superelfml.com": Service.FIRSTMAIL,
    "@syllabicaml.com": Service.FIRSTMAIL,
    "@symmetricalml.com": Service.FIRSTMAIL,
    "@tarosorge.com": Service.FIRSTMAIL,
    "@territorialfml.com": Service.FIRSTMAIL,
    "@trichfml.com": Service.FIRSTMAIL,
    "@tuberfml.com": Service.FIRSTMAIL,
    "@ungainfml.com": Service.FIRSTMAIL,
    "@uterogrml.com": Service.FIRSTMAIL,
    "@uterovesml.com": Service.FIRSTMAIL,
    "@vespertfml.com": Service.FIRSTMAIL,
    "@vocationaml.com": Service.FIRSTMAIL,
    "@volumefml.com": Service.FIRSTMAIL,
    "@xerodeml.com": Service.FIRSTMAIL,
}



import asyncio
from datetime import datetime, timedelta
from typing import Literal
from email import message_from_bytes
from email.utils import parsedate_to_datetime
import re
import pytz
from better_imap.utils import get_service_by_email_address
from better_proxy import Proxy
from .imap import IMAP4_SSL_PROXY
from .errors import IMAPSearchTimeout, IMAPLoginFailed
from .models import EmailMessage, ServiceType
from .utils import extract_email_text


class MailBox:
    DEDAULT_FOLDER_NAMES = ["INBOX", "Junk", "Spam"]

    def __init__(
        self,
        address: str,
        password: str,
        *,
        service: ServiceType = None,
        proxy: Proxy | None = None,
        timeout: float = 10,
        loop: asyncio.AbstractEventLoop = None,
    ):
        if not service:
            service = get_service_by_email_address(address)

        if service.host == "imap.rambler.ru" and "%" in password:
            raise ValueError(
                f"IMAP password contains '%' character. Change your password."
                f" It's a specific rambler.ru error"
            )
        self._address = address
        self._password = password
        self._service = service
        self._connected = False
        self._imap = IMAP4_SSL_PROXY(
            host=service.host,
            proxy=proxy,
            timeout=timeout,
            loop=loop,
        )

    async def __aenter__(self):
        await self.login()
        return self

    async def __aexit__(self, *args):
        await self.logout()

    async def logout(self):
        await self._imap.logout()

    async def login(self):
        if self._connected:
            return
        await self._imap.wait_hello_from_server()
        await self._imap.login(self._address, self._password)
        if self._imap.get_state() == "NONAUTH":
            raise IMAPLoginFailed()
        self._connected = True

    async def _select(self, folder: str):
        try:
            await self._imap.select(mailbox=folder)
            if self._imap.get_state() == "AUTH":
                raise IMAPLoginFailed()
        except TimeoutError:
            raise IMAPLoginFailed("Timeout")

    async def get_message_by_id(self, msg_id) -> EmailMessage:
        typ, msg_data = await self._imap.fetch(msg_id, "(RFC822)")
        if typ == "OK":
            email_bytes = bytes(msg_data[1])
            email_message = message_from_bytes(email_bytes)
            email_sender = email_message.get("from")
            email_receiver = email_message.get("to")
            subject = email_message.get("subject")
            email_date = parsedate_to_datetime(email_message.get("date"))
            if email_date.tzinfo is None:
                email_date = pytz.utc.localize(email_date)
            elif email_date.tzinfo != pytz.utc:
                email_date = email_date.astimezone(pytz.utc)
            message_text = extract_email_text(email_message)
            return EmailMessage(
                text=message_text,
                date=email_date,
                sender=email_sender,
                receiver=email_receiver,
                subject=subject,
            )

    async def fetch_messages(
        self,
        folders: list[str] | None = None,
        *,
        search_criteria: Literal["ALL", "UNSEEN"] = "ALL",
        since: datetime | None = None,
        allowed_senders: list[str] = None,
        allowed_receivers: list[str] = None,
        sender_regex: str | re.Pattern[str] = None,
    ) -> list[EmailMessage]:
        await self.login()
        folders = folders or self.DEDAULT_FOLDER_NAMES
        all_messages = []
        additional_criteria = ""
        if since:
            date_filter = since.strftime("%d-%b-%Y")
            additional_criteria += f" SINCE {date_filter}"
        if allowed_senders:
            senders_criteria = " ".join(f'FROM "{s}"' for s in allowed_senders)
            additional_criteria += f" {senders_criteria}"
        if allowed_receivers:
            receivers_criteria = " ".join(f'TO "{r}"' for r in allowed_receivers)
            additional_criteria += f" {receivers_criteria}"
        for folder in folders:
            await self._select(folder)
            final_search_criteria = f"{search_criteria}{additional_criteria}".strip()
            status, data = await self._imap.search(
                final_search_criteria, charset=self._service.encoding
            )
            if status != "OK" or not data or not data[0]:
                continue
            email_ids = data[0].split()[::-1]
            for e_id_str in email_ids:
                msg_id = e_id_str.decode(self._service.encoding)
                message = await self.get_message_by_id(msg_id)
                if not message:
                    continue
                if since and message.date < since:
                    continue
                if sender_regex and not re.search(
                    sender_regex, message.sender, re.IGNORECASE
                ):
                    continue
                all_messages.append(message)
        return all_messages

    async def search_matches(
        self,
        regex: str | re.Pattern[str],
        folders: list[str] | None = None,
        *,
        search_criteria: Literal["ALL", "UNSEEN"] = "ALL",
        since: datetime = None,
        hours_offset: int | None = None,
        allowed_senders: list[str] = None,
        allowed_receivers: list[str] = None,
        sender_regex: str | re.Pattern[str] = None,
        return_latest: bool = False,
    ) -> str | list[str] | None:
        await self.login()
        if since is None:
            if hours_offset is None:
                since = datetime.now(pytz.utc) - timedelta(hours=hours_offset)
        folders = folders or self.DEDAULT_FOLDER_NAMES
        temp_matches = []
        messages = await self.fetch_messages(
            folders,
            search_criteria=search_criteria,
            since=since,
            allowed_senders=allowed_senders,
            allowed_receivers=allowed_receivers,
            sender_regex=sender_regex,
        )
        for message in messages:
            found = re.findall(regex, message.text)
            if found:
                temp_matches.append((message.date, found[0]))
        temp_matches.sort(key=lambda x: x[0], reverse=True)
        if not temp_matches:
            if return_latest:
                return None
            return []
        if return_latest:
            return temp_matches[0][1]
        return [m[1] for m in temp_matches]

    async def search_match(
        self,
        regex: str | re.Pattern[str],
        folders: list[str] | None = None,
        *,
        search_criteria: Literal["ALL", "UNSEEN"] = "ALL",
        since: datetime = None,
        hours_offset: int = 24,
        allowed_senders: list[str] = None,
        allowed_receivers: list[str] = None,
        sender_regex: str | re.Pattern[str] = None,
    ) -> str | None:
        result = await self.search_matches(
            regex,
            folders=folders,
            search_criteria=search_criteria,
            since=since,
            hours_offset=hours_offset,
            allowed_senders=allowed_senders,
            allowed_receivers=allowed_receivers,
            sender_regex=sender_regex,
            return_latest=True,
        )
        return result

    async def search_with_retry(
        self,
        regex: str | re.Pattern[str],
        folders: list[str] | None = None,
        *,
        allowed_senders: list[str] = None,
        allowed_receivers: list[str] = None,
        sender_email_regex: str | re.Pattern[str] = None,
        since: datetime = None,
        interval: int = 5,
        timeout: int = 90,
    ) -> str | None:
        end_time = asyncio.get_event_loop().time() + timeout
        if since is None:
            since = datetime.now(pytz.utc) - timedelta(seconds=15)
        folders = folders or self.DEDAULT_FOLDER_NAMES
        while asyncio.get_event_loop().time() < end_time:
            match = await self.search_match(
                regex,
                folders=folders,
                allowed_senders=allowed_senders,
                allowed_receivers=allowed_receivers,
                sender_regex=sender_email_regex,
                since=since,
            )
            if match:
                return match
            await asyncio.sleep(interval)
        raise IMAPSearchTimeout(f"No email received within {timeout} seconds")



from email.message import Message
import re
import html2text

from better_imap.errors import UnknownEmailDomain
from better_imap.models import ServiceType
from better_imap.services import DOMAIN_TO_SERVICE

text_maker = html2text.HTML2Text()
text_maker.ignore_links = False


def get_service_by_email_address(email_address: str) -> ServiceType:
    domain = "@" + email_address.split("@")[1]

    if domain not in DOMAIN_TO_SERVICE:
        raise UnknownEmailDomain(email_address)

    return DOMAIN_TO_SERVICE[domain]


def clean_text(text: str) -> str:
    text = text.strip()
    text = re.sub(r"\n\s*\n", "\n", text)
    lines = text.split("\n")
    lines = [line.strip() for line in lines if line.strip()]
    cleaned_text = "\n".join(lines)
    return cleaned_text


def extract_email_text(message: Message):
    text_content = ""
    html_content = ""

    if message.is_multipart():
        for part in message.walk():
            content_type = part.get_content_type()
            content_disposition = str(part.get("Content-Disposition"))

            if "attachment" in content_disposition:
                continue

            if content_type == "text/plain":
                charset = part.get_content_charset() or "utf-8"
                text_content = part.get_payload(decode=True).decode(
                    charset, errors="replace"
                )
            elif content_type == "text/html":
                charset = part.get_content_charset() or "utf-8"
                html_content = part.get_payload(decode=True).decode(
                    charset, errors="replace"
                )
    else:
        content_type = message.get_content_type()
        charset = message.get_content_charset() or "utf-8"

        if content_type == "text/plain":
            text_content = message.get_payload(decode=True).decode(
                charset, errors="replace"
            )
        elif content_type == "text/html":
            html_content = message.get_payload(decode=True).decode(
                charset, errors="replace"
            )

    if text_content and text_content.strip():
        text_content = clean_text(text_content)

    if html_content:
        html_content = text_maker.handle(html_content)

    return text_content + html_content



class BetterImapException(Exception):
    pass


class IMAPLoginFailed(BetterImapException):
    def __init__(self, msg: str = None):
        msg = f". {msg}" if msg else ""
        super().__init__(
            f"IMAP disabled or account banned or incorrect login/password {msg}"
        )


class IMAPSearchTimeout(BetterImapException):
    pass


class UnknownEmailDomain(BetterImapException):
    def __init__(self, email: str):
        super().__init__(f"Unknown email domain {email}")



import asyncio
import ssl

from aioimaplib import IMAP4_SSL, IMAP4, IMAP4ClientProtocol, get_running_loop, Response
from python_socks.async_.asyncio import Proxy as AsyncProxy
from better_proxy import Proxy

IMAP4_SSL_PORT = 993


class IMAP4_SSL_PROXY(IMAP4_SSL):
    def __init__(
            self,
            host: str = "127.0.0.1",
            port: int = IMAP4_SSL_PORT,
            proxy: Proxy | None = None,
            timeout: float = IMAP4.TIMEOUT_SECONDS,
            loop: asyncio.AbstractEventLoop = None,
            ssl_context: ssl.SSLContext = None,
    ):
        self.host = host
        self.port = port
        self.loop = loop
        self.timeout = timeout
        self.ssl_context = ssl_context
        self.proxy = proxy

        self.protocol = None
        self._idle_waiter = None
        self.tasks: set[asyncio.Future] = set()


    async def connect(self) -> None:
        local_loop = self.loop if self.loop else get_running_loop()

        if not self.ssl_context:
            self.ssl_context = ssl.create_default_context()
            self.ssl_context.check_hostname = False
            self.ssl_context.verify_mode = ssl.CERT_NONE

        self.protocol = IMAP4ClientProtocol(local_loop)
        if not self.proxy:
            await self._connect(local_loop)
        else:
            await self._proxy_connect(local_loop)

    async def _connect(self, loop: asyncio.AbstractEventLoop) -> None:
       await loop.create_connection(
            lambda: self.protocol,
            self.host,
            self.port,
            ssl=self.ssl_context,
        )

    async def _proxy_connect(self, loop: asyncio.AbstractEventLoop) -> None:
        async_proxy = AsyncProxy.from_url(self.proxy.as_url)
        sock = await async_proxy.connect(dest_host=self.host, dest_port=self.port)

        await loop.create_connection(
            lambda: self.protocol,
            sock=sock,
            ssl=self.ssl_context,
            server_hostname=self.host,
        )

    async def wait_hello_from_server(self) -> None:
        if not self.protocol:
            await self.connect()
        await super().wait_hello_from_server()

    async def login(self, user: str, password: str) -> Response:
        if not self.protocol:
            await self.connect()
        return await super().login(user, password)
